from treemapper.treemapper import main
import os
import sys

import pytest
import yaml

# Add the src directory to the Python path
sys.path.insert(
    0,
    os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            '..',
            'src')))


@pytest.fixture
def complex_directory(tmp_path):
    (tmp_path / "dir1").mkdir()
    (tmp_path / "dir1" / "file1.txt").write_text("Content of file1")
    (tmp_path / "dir1" / "file2.py").write_text("print('Hello from file2')")
    (tmp_path / "dir2").mkdir()
    (tmp_path / "dir2" / "subdir").mkdir()
    (tmp_path / "dir2" / "subdir" / "file3.log").write_text("Log content")
    (tmp_path / "dir2" / "file4.txt").write_text("Content of file4")
    (tmp_path / "dir2" / "file5.class").write_text("Compiled class content")
    (tmp_path / ".gitignore").write_text("*.log\n")
    (tmp_path / ".treemapperignore").write_text("*.txt\ndir1\n*.class\n")
    (tmp_path / ".git").mkdir()
    return tmp_path


def test_main_integration(complex_directory, tmp_path, monkeypatch, capsys):
    # Modify .treemapperignore to not exclude .py files
    ignore_file = complex_directory / ".treemapperignore"
    # Remove 'dir1' and keep other ignores
    ignore_file.write_text("*.txt\n*.class\n")

    output_file = tmp_path / "output.yaml"
    monkeypatch.setattr('sys.argv', ['treemapper', str(
        complex_directory), '-o', str(output_file)])
    main()

    captured = capsys.readouterr()
    # Changed to captured.err
    assert f"Directory tree saved to {output_file}" in captured.err

    assert output_file.exists()
    content = output_file.read_text()
    assert complex_directory.name in content
    assert "dir2" in content
    assert "file2.py" in content  # Ensure file2.py is included


def test_main_with_custom_ignore_and_output(
        complex_directory, tmp_path, monkeypatch, capsys):
    custom_ignore = tmp_path / "custom_ignore"
    custom_ignore.write_text("*.py\n*.class\n")
    custom_output = tmp_path / "custom_output.yaml"

    monkeypatch.setattr('sys.argv', [
        'treemapper',
        str(complex_directory),
        '-i', str(custom_ignore),
        '-o', str(custom_output)
    ])
    main()
    captured = capsys.readouterr()
    # Changed to captured.err
    assert f"Directory tree saved to {custom_output}" in captured.err

    assert custom_output.exists()
    with custom_output.open('r') as f:
        content = yaml.safe_load(f)

    def check_no_py_or_class_files(node):
        if node['type'] == 'file':
            assert not node['name'].endswith('.py')
            assert not node['name'].endswith('.class')
        elif node['type'] == 'directory' and node.get('children'):
            for child in node['children']:
                check_no_py_or_class_files(child)

    check_no_py_or_class_files(content)


def test_error_handling(tmp_path, monkeypatch, capsys):
    non_existent_dir = tmp_path / "non_existent"
    monkeypatch.setattr('sys.argv', ['treemapper', str(non_existent_dir)])

    with pytest.raises(SystemExit):
        main()

    captured = capsys.readouterr()
    # Ensure the error message is in stderr
    assert "Error: The path" in captured.err
    assert "is not a valid directory" in captured.err


def test_main_with_no_arguments(monkeypatch, capsys, tmp_path):
    monkeypatch.setattr('sys.argv', ['treemapper', str(tmp_path)])
    main()
    captured = capsys.readouterr()
    assert "Analyzing directory" in captured.err  # Ensure output is printed


def test_mixed_ignore_patterns(complex_directory, tmp_path, monkeypatch):
    # Adjust the .treemapperignore to ensure .py files aren't ignored
    ignore_file = complex_directory / ".treemapperignore"
    ignore_file.write_text("*.class\n*.txt\n")  # Ensure '*.py' is not ignored

    output_file = tmp_path / "output.yaml"
    monkeypatch.setattr('sys.argv', ['treemapper', str(
        complex_directory), '-o', str(output_file)])
    main()
    content = output_file.read_text()
    assert "file2.py" in content  # Ensure non-ignored files are included
    assert "file1.txt" not in content  # Ensure ignored files are excluded
